class View:
    """Wrapper for the stdout messages directed to the end-user"""
    def prompt_read(self, field):
        return input("Enter the {}: ".format(field))
    
    def print(self, *args, **kwargs):
        print(*args, sep=" ", end="\r\n", **kwargs)
    
    def tabulate(self, *args):
        if len(args) < 2:
            print(args)
        mid_len = len(args) - 2
        fmt="{left}{mid}{right}".format(left="{:<8}\t", mid=" {} "* mid_len, right="\t{:>10}")
        print(fmt.format(*args))


from . import cipher_utils

import os
import base64
import logging

class Encryptor:
    """BaseClass template for Encryptor class inherited by CipherSuites.
    Essential parameters such as generated keys and cipher modes 
    are dunder-prefixed for reuse while decrypting the message.
    """
    def __init__(self, cipher_params={}, arg_params={}, **kwargs: dict):
        """Initializer for primitives and underlying cipher objects

        :param kwargs: Consolidation of parameters consumed by base cipher library
        :type dict
        """
        param_dict = cipher_params
        param_dict.update(arg_params); param_dict.update(kwargs)
        for key in param_dict:
            setattr(self, "__{}".format(key), param_dict[key])
    
    def encrypt(self, plain_text: str):
        """Template method for encryption of a single message
        
        :param plain_text: input message for encryption
        :type str
        :returns: encrypted output for secure storage
        :rtype str
        """
        return ""

class Decryptor:
    """BaseClass template for Decryptor class inherited by CipherSuites.
    Essential parameters such as generated keys and cipher modes 
    are dunder-prefixed for reuse while decrypting the message.
    """
    def __init__(self, cipher_params={}, arg_params={}, **kwargs: dict):
        """Initializer for primitives and underlying cipher objects

        :param kwargs: Consolidation of parameters consumed by base cipher library
        :type dict
        """
        for key in dict(cipher_params.items() + arg_params.items() + kwargs.items()):
            setattr(self, "__{}".format(key), value)

    def decrypt(self, cipher_text: str):
        """Template method for decryption of a cipher text
        
        :param cipher_text: input message for decryption
        :type str
        :returns: decryption output access/modification
        :rtype str
        """
        return ""

if __name__ != "__main__":
    CIPHER_PKG = {}
    CIPHER_ARGS = {}


import os
import re
import yaml
import logging
import tempfile

class CipherConfig:
    """Wrapper for handling yaml-based parameter configs
    """
    def __init__(self):
        """Propagate config path parameter and default cfg object

        :param cfg_path: Absolute path of YAML config
        :type str
        """
        self._cfg = {}
        self._overwrite_cfg = False

    def load(self, cfg_path):
        """Template method for loading yaml file

        :returns config object associated with cfg_path
        :returntype dict
        """
        if not os.path.isfile(cfg_path):
            raise Exception("unable to find cfg path for loading: {}".format(cfg_path))
        try:
            with open(cfg_path, 'r') as cfg_file:
                cfg = yaml.load(cfg_file, Loader=yaml.FullLoader)
                for param in cfg:
                    self._cfg[param] = {}
                    for key in cfg[param]:
                        if re.match("^.+_bytes$", key) is not None:
                            self._cfg[param][key[:-6]] = cfg[param][key].encode('utf-8')
                        else:
                            self._cfg[param][key] = cfg[param][key]
        except:
            raise Exception("Invalid config file format")
    
    def __call__(self):
        return self._cfg
    
    def _update_cfg(self, config):
        config_serial = {}
        for param in config:
            param_serial = {}
            for field in config[param]:
                if isinstance(config[param][field], bytes):
                    param_serial["{}_bytes".format(field)] = config[param][field].decode('utf-8')
                elif isinstance(config[param][field], int) or \
                     isinstance(config[param][field], float) or \
                     isinstance(config[param][field], str):
                    param_serial[field] = config[param][field]
            config_serial[param] = param_serial
        self._cfg.update(config_serial)
        
    def store(self, cfg_filename, cfg_dirname, **config):
        """Template method for dumping config into yaml file
        """
        _name_prefix, _name_suffix = cfg_filename.rsplit(".")
        self._update_cfg(config)
        cfg_fd, cfg_path = tempfile.mkstemp(
             prefix = _name_prefix + "_", suffix= "." + _name_suffix,
             dir = cfg_dirname, text = True
             )
        try:
            with os.fdopen(cfg_fd, 'w') as cfg_file:
                yaml.dump(self._cfg, cfg_file)
            
            if self.overwrite:
                orig_cfg_path = os.path.join(cfg_dirname, cfg_filename)
                os.rename(cfg_path, orig_cfg_path)
                cfg_path = orig_cfg_path

            return cfg_path
        except:
            raise Exception("Invalid config file format")
    
    @property
    def overwrite(self):
        return self._overwrite_cfg
    
    @overwrite.setter
    def overwrite(self, val):
        if isinstance(val, bool):
            self._overwrite_cfg = val
        else:
            raise Exception("Invalid cfg overwrite option")



import os
import json
import glob
import logging
import traceback
import core.util

class Model(metaclass=core.util.Schema):
    """Represents a data model for a credential record"""
    def __init__(self, model):
        """Initializes a model object for the given schema"""
        self._schema = {}
        self._types = {"int", "float", "str"}
        self.schema_file = core.util.join_path(
                            core.util.get_models_path(), 
                            "{}.json".format(model))
        if not core.util.is_valid_dir(core.util.get_models_path()):
            raise Exception("unable to access schema directory")

    def _validate_schema(self, schema):
        """Validator method for a chosen schema"""
        try:
            model_val = Model.model_template.substitute(**schema)
            for field in schema['fields']:
                field_val = Model.field_template.substitute(**field)
                if field['type'] not in self._types:
                    raise Exception("invalid field type: {}".format(field['type']))
            return schema
        except Exception:
            raise Exception("Invalid schema format provided")

    def load(self):
        """Loads model from schema file"""
        if not core.util.is_valid_file(self.schema_file):
            raise Exception("schema file {} invalid".format(self.schema_file))
        try:
            with open(self.schema_file, 'r') as file_obj:
                self._schema = self._validate_schema(json.load(file_obj))
        except json.decoder.JSONDecodeError:
            raise Exception("invalid json format: " + 
                            self.schema_file)
        except Exception:
            traceback.print_exc()
    
    def get_fields(self):
        if not bool(self._schema):
            self.load()
        return self._schema['fields']
    
    def validate_record(self, content: dict):
        """Validator method for a given record

        :param record content
        :returntype bool
        """
        if not bool(self._schema):
            self.load()
        for field in self._schema['fields']:
            if field['name'] not in content:
                logging.error("required attribute missing: ", field['name'])
            elif not content[field['name']].isascii():
                logging.error(" value of {} is invalid".format(field['name']))
            try:
                res = eval("{}('{}')".format(field['type'], content[field['name']]))
                return True
            except Exception:
                logging.error("Invalid value for {}: {}".format(field['name'], content[field['name']]))
                return False



import json
import traceback
import core.view
import core.record
import ciphers.cipher

class Collection:
    """Houses a collection of records"""
    def __init__(self):
        self._data = []
        self._cur_id = 0
    
    def load_vault(self, vault_content: str, dec: ciphers.cipher.Decryptor):
        res = "[]"
        try:
            res = dec.decrypt(vault_content)
        except TypeError:
            res = dec.decrypt(str(vault_content)[2:-1])
        except Exception:
            traceback.print_exc()
        finally:
            res = res[: res.index("]") + 1]
            fetched_records = json.loads(res)
            self._update_index(fetched_records)
            self._data.extend(fetched_records)
    
    def get_collection(self, enc: ciphers.cipher.Encryptor):
        return enc.encrypt(json.dumps(self._data))

    def get_record(self, record_id: int):
        return self._data[self._calc_record_index(record_id)]

    def add_record(self, rcrd: core.record.Record):
        """Helper for record insertion
        
        :param record: The record object to insert
        :type Record
        """
        if self._cur_id + 1 in map(lambda record: record['record_id'], self._data):
            raise Exception("conflicting record ids present in collection")
        content = {
            "record_id": self._cur_id + 1,
            "record": rcrd()
        }
        self._data.append(content)
        self._cur_id += 1

    def del_record(self, record_id):
        """Helper for record deletion
        
        :param record_id: The identifier of the record to delete
        :type int
        """
        if len(self._data) <= record_id or self._data[record_id]['record_id'] != record_id:
            try:
                tmp_id = [idx for idx, rcrd in 
                    enumerate(self._data) if record_id == rcrd['record_id']][0]
            except Exception:
                raise Exception("Please enter valid record id")
        self._data.pop(tmp_id)

    def modify_record(self, record_id, record):
        """Helper for record modification
        
        :param record_id: The identifier of the record to delete
        :type int
        
        :param record: The record object with modification
        :type Record
        """
        res_id = self._calc_record_index(record_id)
        self._data[res_id]['record'] = record

    def modify_field(self, record_id, field, val):
        """Helper for field modification within a record
        
        :param record_id: The identifier of the record to delete
        :type int
        
        :param field: The field name within the record
        :type str

        :param value: The value of corresponding field
        :type str
        """
        res_id = self._calc_record_index(record_id)
        self._data[res_id]['record'].modify_field(field, val)

    def _calc_record_index(self, record_id):
        if self._data[record_id]['record_id'] != record_id:
            try:
                return [idx for idx, rcrd in 
                    enumerate(self._data) if record_id == rcrd['record_id']][0]
            except IndexError:
                raise Exception("Please enter valid record id")
            except Exception:
                traceback.print_exc()

    def _update_index(self, collection: list):
        if not bool(len(collection)):
            return
        record_ids = set(map(lambda record: record['record_id'], collection))
        if bool(len(record_ids.intersection(set(range(1, self._cur_id + 1))))):
            raise Exception("Record id {} conflicts with another collection record".format(record_id))                   
        elif bool(len(record_ids - (set(range(min(record_ids), max(record_ids) + 1))))):
            raise Exception("Inconsistencies in record ids for the given collection")
        self._cur_id = max(self._cur_id, max(record_ids))
    
    def display(self):
        """Helper for displaying a collection"""
        view_obj = core.view.View()
        view_obj.print()
        view_obj.print("============VAULT CONTENTS==================")
        for record in self._data:
            view_obj.print("#", record['record_id'])
            for field in record['record']:
                view_obj.tabulate(field, record['record'][field])
            view_obj.print("--------------------------------------------")



# https://cryptography.io/en/latest/fernet
import os
import re
import base64
import typing

from cryptography.fernet import Fernet
import cryptography.hazmat.backends.interfaces
import cryptography.hazmat.backends.openssl.backend

from . import cipher, cipher_utils


backend_package = cryptography.hazmat.backends.interfaces.Backend \
if 'Backend' in dir(cryptography.hazmat.backends.interfaces) else \
                cryptography.hazmat.backends.openssl.backend

class Encryptor(cipher.Encryptor):
    """Encryptor class for symmetric encryption through Fernet module.
    Essential parameters such as generated keys and cipher modes 
    are dunder-prefixed for reuse while decrypting the message.
    """
    def __init__(self, **kwargs):
        """Initializer for primitives and underlying cipher objects

        :param kwargs: Consolidation of parameters consumed by base cipher library
        :type dict
        """
        backend = None
        cipher_params = kwargs.get("cipher_params", {})
        backend_val = cipher_params.pop("backend", None)
        if backend_val is not None:
            try:
                if backend_val not in CIPHER_ARGS['backend']:
                    raise Exception("")
                backend = vars(backend)[backend_val]
            except:
                raise Exception("backend {} not supported by Fernet".format(backend_val))
        self._key = cipher_params.get("key", Fernet.generate_key())
        self.__encoding = kwargs.get("arg_params", {}).get("encoding", "utf-8")
        self.fernet = Fernet(self._key, backend, **kwargs.get('cipher_params',{}))

    def encrypt(self, plain_text: str):
        """Template method for encryption of a single message
        
        :param plain_text: input message for encryption
        :type str
        :returns: encrypted output for secure storage
        :rtype str
        """
        return self.fernet.encrypt(plain_text.encode(self.__encoding))

class Decryptor(cipher.Decryptor):
    """Decryptor class for symmetric decryption through Fernet module.
    Essential parameters such as generated keys and cipher modes 
    are dunder-prefixed for reuse while decrypting the message.
    """
    def __init__(self, **kwargs):
        """Initializer for primitives and underlying cipher objects

        :param kwargs: Consolidation of parameters consumed by base cipher library
        :type dict
        """
        backend = None
        cipher_params = kwargs.get("cipher_params", {})
        backend_val = cipher_params.pop("backend", None)
        if "key" not in cipher_params:
            raise Exception("Fernet decryption requires a symmetric key as input")
        if backend_val is not None:
            try:
                if backend_val not in CIPHER_ARGS['backend']:
                    raise Exception("")
                backend = vars(backend)[backend_val]
            except:
                raise Exception("backend {} not supported by Fernet".format(backend_val))
        self._key = cipher_params.pop("key")
        self.__encoding = kwargs.get("arg_params", {}).get("encoding", "utf-8")
        self.fernet = Fernet(self._key, backend, **cipher_params)

    def decrypt(self, cipher_text):
        """Template method for decryption of a cipher text
        
        :param cipher_text: input message for decryption
        :type str
        :returns: decryption output access/modification
        :rtype str
        """
        return self.fernet.decrypt(cipher_text).decode(self.__encoding)

if __name__ != "__main__":
    CIPHER_PKG = {'backend': backend_package}
    CIPHER_ARGS = {k:cipher_utils.expand_attrs(v) \
                    for k,v in CIPHER_PKG.items()}
